with Grt.Errors; use Grt.Errors;
with Grt.Stdio; use Grt.Stdio;
with GNAT.Table;
with System; use System;
with Ada.Unchecked_Conversion;

package body Grt.Files is
   subtype C_Files is Grt.Stdio.FILEs;

   type File_Signature_Type is record
      Len : Natural;
      Max_Len : Natural;
      Str : String_Ptr;
   end record;

   type File_Signature_Acc is access File_Signature_Type;

   function To_File_Signature_Acc is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => File_Signature_Acc);
   function To_Address is new Ada.Unchecked_Conversion
     (Source => File_Signature_Acc, Target => System.Address);

   type File_Entry_Type is record
      Stream : C_Files;
      Is_Text : Boolean;
      Signature : File_Signature_Acc;
   end record;

   Nul : constant Character := Character'Val (0);
   Nl : constant Character := Character'Val (10);  -- nl or '\n'.

   package Files_Table is new GNAT.Table
     (Table_Component_Type => File_Entry_Type,
      Table_Index_Type => Ghdl_File_Index,
      Table_Low_Bound => 1,
      Table_Initial => 2,
      Table_Increment => 100);

   function Create_File (Is_Text : Boolean; Signature : File_Signature_Acc)
                        return Ghdl_File_Index
   is
      Res : Ghdl_File_Index;
   begin
      Files_Table.Increment_Last;
      Res := Files_Table.Last;
      Files_Table.Table (Res) := (Stream => NULL_Stream,
                                  Is_Text => Is_Text,
                                  Signature => Signature);
      return Res;
   end Create_File;

   procedure File_Error (File : Ghdl_File_Index)
   is
      pragma Unreferenced (File);
   begin
      Internal_Error ("file: IO error");
   end File_Error;

   function Ghdl_Text_File_Elaborate return Ghdl_File_Index is
   begin
      return Create_File (True, null);
   end Ghdl_Text_File_Elaborate;

   function Ghdl_File_Elaborate (Sig : System.Address) return Ghdl_File_Index
   is
   begin
      return Create_File (False, To_File_Signature_Acc (Sig));
   end Ghdl_File_Elaborate;

   function Get_File (Index : Ghdl_File_Index) return C_Files
   is
   begin
      if Index not in Files_Table.First .. Files_Table.Last then
         Internal_Error ("get_file: bad file index");
      end if;
      return Files_Table.Table (Index).Stream;
   end Get_File;

   procedure Check_File_Mode (Index : Ghdl_File_Index; Is_Text : Boolean)
   is
   begin
      if Files_Table.Table (Index).Is_Text /= Is_Text then
         Internal_Error ("check_file_mode: bad file mode");
      end if;
   end Check_File_Mode;

   function Ghdl_File_Endfile (File : Ghdl_File_Index) return Boolean
   is
      Stream : C_Files;
      C : int;
   begin
      Stream := Get_File (File);
      if feof (Stream) /= 0 then
         return True;
      end if;
      C := fgetc (Stream);
      if C < 0 then
         return True;
      end if;
      if ungetc (C, Stream) /= C then
         Error ("internal error: ungetc");
      end if;
      return False;
   end Ghdl_File_Endfile;

   Sig_Header : constant String := "#GHDL-BINARY-FILE-0.0" & Nl;

   function File_Open (File : Ghdl_File_Index;
                       Mode : Ghdl_Integer;
                       Str : Std_String_Ptr)
     return Ghdl_Integer
   is
      Name : String (1 .. Integer (Str.Bounds.Dim_1.Length) + 1);
      Str_Mode : String (1 .. 3);
      F : C_Files;
      Sig : File_Signature_Acc;
   begin
      F := Get_File (File);

      if F /= NULL_Stream then
         --  File was already open.
         return Status_Error;
      end if;

      --  Copy file name and convert it to a C string (NUL terminated).
      for I in 0 .. Str.Bounds.Dim_1.Length - 1 loop
         Name (1 + Natural (I)) := Str.Base (I);
      end loop;
      Name (Name'Last) := Nul;

      if Name = "STD_INPUT" & Nul then
         if Mode /= Read_Mode then
            return Mode_Error;
         end if;
         F := stdin;
      elsif Name = "STD_OUTPUT" & Nul then
         if Mode /= Write_Mode then
            return Mode_Error;
         end if;
         F := stdout;
      else
         case Mode is
            when Read_Mode =>
               Str_Mode (1) := 'r';
            when Write_Mode =>
               Str_Mode (1) := 'w';
            when Append_Mode =>
               Str_Mode (1) := 'a';
            when others =>
               --  Bad mode, cannot happen.
               Internal_Error ("file_open: bad open mode");
         end case;
         if Files_Table.Table (File).Is_Text then
            Str_Mode (2) := Nul;
         else
            Str_Mode (2) := 'b';
            Str_Mode (3) := Nul;
         end if;
         F := fopen (Name'Address, Str_Mode'Address);
         if F = NULL_Stream then
            return Name_Error;
         end if;
      end if;
      Sig := Files_Table.Table (File).Signature;
      if Sig /= null then
         case Mode is
            when Write_Mode =>
               if fwrite (Sig_Header'Address, 1, Sig_Header'Length, F)
                 /= Sig_Header'Length
               then
                  File_Error (File);
               end if;
               if fwrite (Sig.Str (1)'Address, 1, size_t (Sig.Len), F)
                 /= size_t (Sig.Len)
               then
                  File_Error (File);
               end if;
            when Read_Mode =>
               declare
                  Hdr : String (1 .. Sig_Header'Length);
                  Sig_Buf : String (1 .. Sig.Len);
               begin
                  if fread (Hdr'Address, 1, Hdr'Length, F) /= Hdr'Length then
                     File_Error (File);
                  end if;
                  if Hdr /= Sig_Header then
                     File_Error (File);
                  end if;
                  if fread (Sig_Buf'Address, 1, Sig_Buf'Length, F)
                    /= Sig_Buf'Length
                  then
                     File_Error (File);
                  end if;
                  if Sig_Buf /= Sig.Str (1 .. Sig.Len) then
                     File_Error (File);
                  end if;
               end;
            when Append_Mode =>
               null;
            when others =>
               null;
         end case;
      end if;
      Files_Table.Table (File).Stream := F;
      return Open_Ok;
   end File_Open;

   procedure Ghdl_Text_File_Open
     (File : Ghdl_File_Index; Mode : Ghdl_Integer; Str : Std_String_Ptr)
   is
      Res : Ghdl_Integer;
   begin
      Check_File_Mode (File, True);

      Res := File_Open (File, Mode, Str);

      if Res /= Open_Ok then
         Error_C ("open: cannot open text file ");
         Error_E (String (Str.Base (0 .. Str.Bounds.Dim_1.Length - 1)));
      end if;
   end Ghdl_Text_File_Open;

   procedure Ghdl_File_Open
     (File : Ghdl_File_Index; Mode : Ghdl_Integer; Str : Std_String_Ptr)
   is
      Res : Ghdl_Integer;
   begin
      Check_File_Mode (File, False);

      Res := File_Open (File, Mode, Str);

      if Res /= Open_Ok then
         Error_C ("open: cannot open file ");
         Error_E (String (Str.Base (0 .. Str.Bounds.Dim_1.Length - 1)));
      end if;
   end Ghdl_File_Open;

   function Ghdl_Text_File_Open_Status
     (File : Ghdl_File_Index; Mode : Ghdl_Integer; Str : Std_String_Ptr)
     return Ghdl_Integer
   is
   begin
      Check_File_Mode (File, True);
      return File_Open (File, Mode, Str);
   end Ghdl_Text_File_Open_Status;

   function Ghdl_File_Open_Status
     (File : Ghdl_File_Index; Mode : Ghdl_Integer; Str : Std_String_Ptr)
     return Ghdl_Integer
   is
   begin
      Check_File_Mode (File, False);
      return File_Open (File, Mode, Str);
   end Ghdl_File_Open_Status;

   procedure Ghdl_Text_Write (File : Ghdl_File_Index; Str : Std_String_Ptr)
   is
      Res : C_Files;
      R : size_t;
      R1 : int;
   begin
      Res := Get_File (File);
      Check_File_Mode (File, True);
      if Res = NULL_Stream then
         Error ("write to a non-opened file");
      end if;
      --  FIXME: check mode.
      R := fwrite (Str.Base (0)'Address,
                   size_t (Str.Bounds.Dim_1.Length), 1, Res);
      --  FIXME: check r
      --  Write '\n'.
      R1 := fputc (Character'Pos (Nl), Res);
      R1 := fflush (Res);
   end Ghdl_Text_Write;

   procedure Ghdl_Write_Scalar (File : Ghdl_File_Index;
                                Ptr : Ghdl_Ptr;
                                Length : Ghdl_Index_Type)
   is
      Res : C_Files;
      R : size_t;
      R1 : int;
   begin
      Res := Get_File (File);
      Check_File_Mode (File, False);
      if Res = NULL_Stream then
         Error ("write to a non-opened file");
      end if;
      --  FIXME: check mode.
      R := fwrite (System.Address (Ptr), size_t (Length), 1, Res);
      if R /= 1 then
         Error ("write_scalar failed");
      end if;
      R1 := fflush (Res);
   end Ghdl_Write_Scalar;

   procedure Ghdl_Read_Scalar (File : Ghdl_File_Index;
                               Ptr : Ghdl_Ptr;
                               Length : Ghdl_Index_Type)
   is
      Res : C_Files;
      R : size_t;
   begin
      Res := Get_File (File);
      Check_File_Mode (File, False);
      if Res = NULL_Stream then
         Error ("write to a non-opened file");
      end if;
      --  FIXME: check mode.
      R := fread (System.Address (Ptr), size_t (Length), 1, Res);
      if R /= 1 then
         Error ("read_scalar failed");
      end if;
   end Ghdl_Read_Scalar;

   function Ghdl_Text_Read_Length (File : Ghdl_File_Index;
                                   Str : Std_String_Ptr)
     return Std_Integer
   is
      Stream : C_Files;
      Len : int;
      Idx : Ghdl_Index_Type;
   begin
      Stream := Get_File (File);
      Check_File_Mode (File, True);
      Len := int (Str.Bounds.Dim_1.Length);
      if fgets (Str.Base (0)'Address, Len, Stream) = Null_Address then
         Internal_Error ("ghdl_text_read_length: fgets error");
      end if;
      --  Compute the length.
      for I in Ghdl_Index_Type loop
         if Str.Base (I) = Nul then
            Idx := I;
            exit;
         end if;
      end loop;
      if Idx > 0 and then Str.Base (Idx - 1) = Nl then
         Idx := Idx - 1;
      end if;
      return Std_Integer (Idx);
   end Ghdl_Text_Read_Length;

   procedure Ghdl_Text_File_Close (File : Ghdl_File_Index)
   is
      Stream : C_Files;
   begin
      Stream := Get_File (File);
      Check_File_Mode (File, True);
      if fclose (Stream) /= 0 then
         Internal_Error ("ghdl_text_file_close: fclose error");
      end if;
      Files_Table.Table (File).Stream := NULL_Stream;
   end Ghdl_Text_File_Close;


   procedure File_Signature_Add (Fs : File_Signature_Acc; Str : String)
   is
      function malloc (Size : Natural) return String_Ptr;
      pragma Import (C, malloc);

      function realloc (Ptr : String_Ptr; Size : Natural) return String_Ptr;
      pragma Import (C, realloc);

      Nlen : Natural;
   begin
      Nlen := Fs.Len + Str'Length;
      if Nlen > Fs.Max_Len then
         if Fs.Len = 0 then
            Fs.Max_Len := 32;
            Fs.Str := malloc (Fs.Max_Len);
         else
            while Fs.Max_Len < Nlen loop
               Fs.Max_Len := Fs.Max_Len * 2;
            end loop;
            Fs.Str := realloc (Fs.Str, Fs.Max_Len);
         end if;
      end if;
      for I in Str'Range loop
         Fs.Str (Fs.Len + 1) := Str (I);
         Fs.Len := Fs.Len + 1;
      end loop;
   end File_Signature_Add;

   procedure File_Signature_Add (Fs: File_Signature_Acc; Td : Ghdl_Desc_Ptr)
   is
   begin
      case Td.Kind is
         when Ghdl_Type_B2 =>
            File_Signature_Add (Fs, "b");
         when Ghdl_Type_E8
           | Ghdl_Subtype_E8 =>
            File_Signature_Add (Fs, "e");
         when Ghdl_Type_E32 =>
            File_Signature_Add (Fs, "E");
         when Ghdl_Type_I32
           | Ghdl_Subtype_I32 =>
            File_Signature_Add (Fs, "i");
         when Ghdl_Type_I64 =>
            File_Signature_Add (Fs, "I");
         when Ghdl_Type_F64
           | Ghdl_Subtype_F64 =>
            File_Signature_Add (Fs, "f");
         when Ghdl_Type_P64
           | Ghdl_Subtype_P64 =>
            File_Signature_Add (Fs, "p");
         when Ghdl_Type_Record =>
            File_Signature_Add (Fs, "<");
            for I in 0 .. Td.Rec.Nbr_El - 1 loop
               File_Signature_Add (Fs, Td.Rec.Els (I));
            end loop;
            File_Signature_Add (Fs, ">");
         when Ghdl_Subtype_Array =>
            declare
               Bt : Ghdl_Desc_Ptr;
            begin
               Bt := Td.Sub_Arr.Basetype;
               File_Signature_Add (Fs, "[");
               for I in 0 .. Bt.Arr.Nbr_Dim - 1 loop
                  File_Signature_Add (Fs, Bt.Arr.Indexes (I));
                  --  FIXME: length
               end loop;
               File_Signature_Add (Fs, "]");
               File_Signature_Add (Fs, Bt.Arr.El_Desc);
            end;
         when Ghdl_Type_Access
           | Ghdl_Type_Array
           | Ghdl_Type_Any =>
            Internal_Error ("file_signature_add");
      end case;
   end File_Signature_Add;

   function Ghdl_File_Create_Signature (Td : Ghdl_Desc_Ptr)
                                       return System.Address
   is
      Res : File_Signature_Acc;
   begin
      Res := new File_Signature_Type'(Len => 0, Max_Len => 0, Str => null);

      case Td.Kind is
         when Ghdl_Type_Array =>
            File_Signature_Add (Res, "[]");
            File_Signature_Add (Res, Td.Arr.El_Desc);
         when others =>
            File_Signature_Add (Res, Td);
      end case;
      File_Signature_Add (Res, "." & Nl);
      return To_Address (Res);
   end Ghdl_File_Create_Signature;
end Grt.Files;

