with Grt.Errors; use Grt.Errors;
with Ada.Unchecked_Conversion;
with System; use System;
with Grt.Processes; use Grt.Processes;
with Grt.Typedesc; use Grt.Typedesc;

package body Grt.Names is
   function To_Str_String_Boundp is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Std_String_Boundp);

   function To_Std_String_Basep is new Ada.Unchecked_Conversion
     (Source => String_Ptr, Target => Std_String_Basep);

   function To_Std_String_Basep is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Std_String_Basep);

   procedure Get_Name (Res : Std_String_Ptr;
                       Instance : Ghdl_Instance_Name_Acc;
                       Name : Ghdl_Str_Len_Ptr;
                       Is_Path : Boolean)
   is
      Buf : String (1 .. 128);
      Buf_Len : Natural := 0;

      procedure Append (Str : String) is
      begin
         Buf (Buf_Len + 1 .. Buf_Len + Str'Length) := Str;
         Buf_Len := Buf_Len + Str'Length;
      end Append;

      generic
         type Ntype is range <>;
         Max_Len : Natural;
      procedure Add_Ntype (N : Ntype);

      procedure Add_Ntype (N : Ntype)
      is
         Str : String (1 .. Max_Len);
         P : Natural := Str'Last;
         V : Ntype;
      begin
         if N > 0 then
            V := -N;
         else
            V := N;
         end if;
         loop
            Str (P) := Character'Val (48 - (V rem 10));
            V := V / 10;
            exit when V = 0;
            P := P - 1;
         end loop;
         if N < 0 then
            P := P - 1;
            Str (P) := '-';
         end if;
         Append (Str (P .. Max_Len));
      end Add_Ntype;

      procedure Add_I32 is new Add_Ntype (Ntype => Ghdl_I32, Max_Len => 11);

      procedure Add_Name (Name : Ghdl_Str_Len_Ptr)
      is
      begin
         Append (Name.Str (1 .. Name.Len));
      end Add_Name;

      procedure Add_Instance (Inst : Ghdl_Instance_Name_Acc)
      is
      begin
         case Inst.Kind is
            when Ghdl_Name_Block =>
               Add_Instance (Inst.Parent);
               Add_Name (Inst.Name);
               Append (":");
            when Ghdl_Name_Instance =>
               Add_Instance (Inst.Parent);
               Add_Name (Inst.Name);
               if Is_Path then
                  Append (":");
               else
                  Append ("@");
               end if;
            when Ghdl_Name_Architecture =>
               --  The entity.
               Add_Instance (Inst.Parent.Parent);
               if not Is_Path then
                  Add_Name (Inst.Parent.Name);
                  Append ("(");
                  Add_Name (Inst.Name);
                  Append ("):");
               end if;
            when Ghdl_Name_Entity =>
               Add_Instance (Inst.Parent);
               if not Is_Path then
                  Add_Name (Inst.Name);
                  Append ("(");
                  declare
                     Child : Ghdl_Instance_Name_Acc;
                  begin
                     Child := Inst.Child;
                     while Child /= null loop
                        if Child.Kind = Ghdl_Name_Architecture then
                           Add_Name (Child.Name);
                           exit;
                        else
                           Child := Child.Brother;
                        end if;
                     end loop;
                  end;
                  Append ("):");
               end if;
            when Ghdl_Name_Top =>
               Append (":");
               if Is_Path then
                  Add_Name (Inst.Child.Name);
                  Append (":");
               end if;
            when Ghdl_Name_Generate_Iterative =>
               Add_Instance (Inst.Parent);
               Add_Name (Inst.Name);
               Append ("(");
               case Inst.Iter_Typedesc.Kind is
                  when Ghdl_Type_I32 =>
                     Add_I32 (Inst.Iter_Addr.I32);
                  when others =>
                     Internal_Error ("get_name/add_instance(2)");
               end case;
               Append ("):");
            when others =>
               Internal_Error ("get_name/add_instance");
         end case;
      end Add_Instance;

      procedure Memcpy (Dst : Address; Src : Address; Len : Integer);
      pragma Import (C, Memcpy);

      Bounds : Std_String_Boundp;
      Len : Natural;
   begin
      if Instance /= null then
         Add_Instance (Instance);
         Len := Buf_Len + Name.Len;
      else
         Len := Name.Len;
      end if;

      Bounds := To_Str_String_Boundp
        (Ghdl_Stack2_Allocate (Std_String_Bound'Size / System.Storage_Unit));
      Bounds.Dim_1.Left := 1;
      Bounds.Dim_1.Right := Ghdl_Integer (Len);
      Bounds.Dim_1.Dir := Dir_To;
      Bounds.Dim_1.Length := Ghdl_Index_Type (Len);
      Res.Bounds := Bounds;
      if Instance /= null then
         Res.Base := To_Std_String_Basep
           (Ghdl_Stack2_Allocate (Ghdl_Index_Type (Len)));
         Memcpy (Res.Base (0)'Address, Buf'Address, Buf_Len);
         Memcpy (Res.Base (Ghdl_Index_Type (Buf_Len))'Address,
                 Name.Str (1)'Address,
                 Name.Len);
      else
         Res.Base := To_Std_String_Basep (Name.Str);
      end if;
   end Get_Name;

   procedure Ghdl_Get_Path_Name (Res : Std_String_Ptr;
                                 Instance : Ghdl_Instance_Name_Acc;
                                 Name : Ghdl_Str_Len_Ptr)
   is
   begin
      Get_Name (Res, Instance, Name, True);
   end Ghdl_Get_Path_Name;

   procedure Ghdl_Get_Instance_Name (Res : Std_String_Ptr;
                                     Instance : Ghdl_Instance_Name_Acc;
                                     Name : Ghdl_Str_Len_Ptr)
   is
   begin
      Get_Name (Res, Instance, Name, False);
   end Ghdl_Get_Instance_Name;

end Grt.Names;
