with GNAT.Table;
with Ada.Unchecked_Conversion;
with Ada.Unchecked_Deallocation;
with Grt.Stack2; use Grt.Stack2;
with Grt.Disp;
with Grt.Signals; use Grt.Signals;
with Grt.Errors; use Grt.Errors;
with Grt.Stacks; use Grt.Stacks;
with Grt.Options;
with Grt.Hierarchy;
with Grt.Vcd;

package body Grt.Processes is
   --  Access to a process subprogram.
   type Proc_Acc is access procedure (Self : System.Address);

   --  Simply linked list for sensitivity.
   type Sensitivity_El;
   type Sensitivity_Acc is access Sensitivity_El;
   type Sensitivity_El is record
      Sig : Ghdl_Signal_Ptr;
      Next : Sensitivity_Acc;
   end record;

   Last_Time : Std_Time := Std_Time'Last;

   --  State of a process.
   type Process_State is
     (
      --  Sensitized process.  Its state cannot change.
      State_Sensitized,

      --  Non-sensitized process being suspended.
      State_Wait,

      --  Non-sensitized process being awaked by a wait timeout.  This state
      --  is transcient.
      State_Timeout,

      --  Non-sensitized process waiting until end.
      State_Dead);

   type Process_Type is record
      --  Stack for the process.
      --  This must be the first field of the record (and this is the only
      --  part visible).
      --  Must be NULL_STACK for sensitized processes.
      Stack : Stack_Type;

      --  Subprogram containing process code.
      Subprg : Proc_Acc;

      --  Instance (THIS parameter) for the subprogram.
      This : System.Address;

      --  Name of the process.
      Name : Grt.Hierarchy.Ghdl_Instance_Name_Acc;

      --  True if the process is resumed and will be run at next cycle.
      Resumed : Boolean;

      --  True if the process is postponed.
      Postponed : Boolean;

      State : Process_State;

      --  Timeout value for wait.
      Timeout : Std_Time;

      --  Sensitivity list.
      Sensitivity : Sensitivity_Acc;
   end record;
   type Process_Acc is access all Process_Type;

   --  Per 'thread' data.
   --  The process being executed.
   Cur_Proc_Id : Process_Id;

   Cur_Proc : Process_Acc;
   pragma Export (C, Cur_Proc, "grt_cur_proc");

   --  The secondary stack for the thread.
   Stack2 : Stack2_Ptr;

   package Process_Table is new GNAT.Table
     (Table_Component_Type => Process_Type,
      Table_Index_Type => Process_Id,
      Table_Low_Bound => 1,
      Table_Initial => 1,
      Table_Increment => 100);

   procedure Free is new Ada.Unchecked_Deallocation
     (Name => Sensitivity_Acc, Object => Sensitivity_El);

   procedure Init is
   begin
      Process_Table.Init;
   end Init;

   function Get_Current_Process_Id return Process_Id
   is
   begin
      return Cur_Proc_Id;
   end Get_Current_Process_Id;

   procedure Process_Register (This : System.Address;
                               Proc : System.Address;
                               Name : Ghdl_Instance_Name_Acc;
                               State : Process_State;
                               Postponed : Boolean)
   is
      use Grt.Hierarchy;

      function To_Proc_Acc is new Ada.Unchecked_Conversion
        (Source => System.Address, Target => Proc_Acc);
      Stack : Stack_Type;
   begin
      if State /= State_Sensitized then
         Stack := Stack_Create (Proc, This);
      else
         Stack := Null_Stack;
      end if;
      Process_Table.Increment_Last;
      Process_Table.Table (Process_Table.Last) :=
        (Subprg => To_Proc_Acc (Proc),
         This => This,
         Name => Name,
         Sensitivity => null,
         Resumed => True,
         Postponed => Postponed,
         State => State,
         Timeout => Bad_Time,
         Stack => Stack);
      --  Used to create drivers.
      Cur_Proc_Id := Process_Table.Last;
   end Process_Register;

   procedure Ghdl_Process_Register
     (Instance : System.Address;
      Proc : System.Address;
      Name : Ghdl_Instance_Name_Acc)
   is
   begin
      Process_Register (Instance, Proc, Name, State_Timeout, False);
   end Ghdl_Process_Register;

   procedure Ghdl_Sensitized_Process_Register
     (Instance : System.Address;
      Proc : System.Address;
      Name : Ghdl_Instance_Name_Acc)
   is
   begin
      Process_Register (Instance, Proc, Name, State_Sensitized, False);
   end Ghdl_Sensitized_Process_Register;

   procedure Ghdl_Postponed_Process_Register
     (Instance : System.Address;
      Proc : System.Address;
      Name : Ghdl_Instance_Name_Acc)
   is
   begin
      Process_Register (Instance, Proc, Name, State_Timeout, True);
   end Ghdl_Postponed_Process_Register;

   procedure Ghdl_Postponed_Sensitized_Process_Register
     (Instance : System.Address;
      Proc : System.Address;
      Name : Ghdl_Instance_Name_Acc)
   is
   begin
      Process_Register (Instance, Proc, Name, State_Sensitized, True);
   end Ghdl_Postponed_Sensitized_Process_Register;

   procedure Ghdl_Process_Add_Sensitivity (Sig : Ghdl_Signal_Ptr)
   is
   begin
      Resume_Process_If_Event (Sig, Process_Table.Last);
   end Ghdl_Process_Add_Sensitivity;

      procedure Resume_Process (Proc : Process_Id) is
   begin
      Process_Table.Table (Proc).Resumed := True;
   end Resume_Process;

   function Ghdl_Stack2_Allocate (Size : Ghdl_Index_Type)
     return System.Address
   is
   begin
      return Grt.Stack2.Allocate (Stack2, Size);
   end Ghdl_Stack2_Allocate;

   function Ghdl_Stack2_Mark return Mark_Id is
   begin
      if Stack2 = Null_Stack2_Ptr then
         Stack2 := Grt.Stack2.Create;
      end if;
      return Grt.Stack2.Mark (Stack2);
   end Ghdl_Stack2_Mark;

   procedure Ghdl_Stack2_Release (Mark : Mark_Id) is
   begin
      Grt.Stack2.Release (Stack2, Mark);
   end Ghdl_Stack2_Release;

   function To_Acc is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Process_Acc);

   procedure Ghdl_Process_Wait_Add_Sensitivity (Sig : Ghdl_Signal_Ptr)
   is
      El : Sensitivity_Acc;
   begin
      El := new Sensitivity_El'(Sig => Sig,
                                Next => Cur_Proc.Sensitivity);
      Cur_Proc.Sensitivity := El;
   end Ghdl_Process_Wait_Add_Sensitivity;

   procedure Ghdl_Process_Wait_Set_Timeout (Time : Std_Time)
   is
   begin
      if Time < 0 then
         --  LRM93 8.1
         Error ("negative timeout clause");
      end if;
      Cur_Proc.Timeout := Current_Time + Time;
   end Ghdl_Process_Wait_Set_Timeout;

   function Ghdl_Process_Wait_Suspend return Boolean is
   begin
      if Cur_Proc.State = State_Sensitized then
         Error ("wait statement in a sensitized process");
      end if;
      --  Suspend this process.
      Cur_Proc.State := State_Wait;
      Stack_Switch (Main_Stack, Cur_Proc.Stack);
      return Cur_Proc.State = State_Timeout;
   end Ghdl_Process_Wait_Suspend;

   procedure Ghdl_Process_Wait_Close
   is
      El : Sensitivity_Acc;
      N_El : Sensitivity_Acc;
   begin
      El := Cur_Proc.Sensitivity;
      Cur_Proc.Sensitivity := null;
      while El /= null loop
         N_El := El.Next;
         Free (El);
         El := N_El;
      end loop;
   end Ghdl_Process_Wait_Close;

   procedure Ghdl_Process_Wait_Exit
   is
   begin
      if Cur_Proc.State = State_Sensitized then
         Error ("wait statement in a sensitized process");
      end if;
      --  Mark this process as dead, in order to kill it.
      --  It cannot be killed now, since this code is still in the process.
      Cur_Proc.State := State_Dead;
      --  Suspend this process.
      Stack_Switch (Main_Stack, Cur_Proc.Stack);
   end Ghdl_Process_Wait_Exit;

   procedure Ghdl_Process_Wait_Timeout (Time : Std_Time)
   is
   begin
      if Cur_Proc.State = State_Sensitized then
         Error ("wait statement in a sensitized process");
      end if;
      if Time < 0 then
         --  LRM93 8.1
         Error ("negative timeout clause");
      end if;
      Cur_Proc.Timeout := Current_Time + Time;
      Cur_Proc.State := State_Wait;
      --  Suspend this process.
      Stack_Switch (Main_Stack, Cur_Proc.Stack);
   end Ghdl_Process_Wait_Timeout;

   function Compute_Next_Time return Std_Time
   is
      Res : Std_Time;
   begin
      --  f) The time of the next simulation cycle, Tn, is determined by
      --     setting it to the earliest of
      --     1) TIME'HIGH
      Res := Std_Time'Last;

      --     2) The next time at which a driver becomes active, or
      Res := Std_Time'Min (Res, Grt.Signals.Find_Next_Time);

      if Res = Current_Time then
         return Res;
      end if;

      --     3) The next time at which a process resumes.
      for I in Process_Table.First .. Process_Table.Last loop
         declare
            Proc : Process_Type renames Process_Table.Table (I);
         begin
            if Proc.State = State_Wait
              and then Proc.Timeout >= Current_Time
              and then Proc.Timeout < Res
            then
               if Proc.Timeout = Current_Time then
                  return Current_Time;
               end if;
               Res := Proc.Timeout;
            end if;
         end;
      end loop;

      return Res;
   end Compute_Next_Time;

   --  Run resumed processes.
   --  If POSTPONED is true, resume postponed processes, else resume
   --  non-posponed processes.
   --  HAS_RUN will be set to true if a process has been run.
   procedure Run_Processes (Postponed : Boolean; Has_Run : out Boolean)
   is
   begin
      Has_Run := False;

      for I in Process_Table.First .. Process_Table.Last loop
         if Process_Table.Table (I).Postponed = Postponed
           and Process_Table.Table (I).Resumed
         then
            if Grt.Options.Trace_Processes then
               Grt.Disp.Put ("run process ");
               Grt.Disp.Disp_Instance_Name (Process_Table.Table (I).Name);
               Grt.Disp.New_Line;
            end if;
            Process_Table.Table (I).Resumed := False;
            Has_Run := True;
            Cur_Proc_Id := I;
            Cur_Proc := To_Acc (Process_Table.Table (I)'Address);
            if Cur_Proc.State = State_Sensitized then
               Cur_Proc.Subprg.all (Cur_Proc.This);
            else
               Stack_Switch (Cur_Proc.Stack, Main_Stack);
            end if;
            if Grt.Options.Checks then
               Ghdl_Signal_Internal_Checks;
               Grt.Stack2.Check_Empty (Stack2);
            end if;
         end if;
      end loop;
   end Run_Processes;

   procedure Initialization_Phase
   is
      Has_Run : Boolean;
   begin
      --  LRM93 12.6.4
      --  At the beginning of initialization, the current time, Tc, is assumed
      --  to be 0 ns.
      Current_Time := 0;

      --  The initialization phase consists of the following steps:
      --  - The driving value and the effective value of each explicitly
      --    declared signal are computed, and the current value of the signal
      --    is set to the effective value.  This value is assumed to have been
      --    the value of the signal for an infinite length of time prior to
      --    the start of the simulation.
      Init_Signals;

      --  - The value of each implicit signal of the form S'Stable(T) or
      --    S'Quiet(T) is set to true.  The value of each implicit signal of
      --    the form S'Delayed is set to the initial value of its prefix, S.
      --  GHDL: already done when the signals are created.
      null;

      --  - The value of each implicit GUARD signal is set to the result of
      --    evaluating the corresponding guard expression.
      null;

      --  - Each nonpostponed process in the model is executed until it
      --    suspends.
      Run_Processes (Postponed => False, Has_Run => Has_Run);

      --  - Each postponed process in the model is executed until it suspends.
      Run_Processes (Postponed => True, Has_Run => Has_Run);

      --  - The time of the next simulation cycle (which in this case is the
      --    first simulation cycle), Tn, is calculated according to the rules
      --    of step f of the simulation cycle, below.
      Current_Time := Compute_Next_Time;
   end Initialization_Phase;

   --  Launch a simulation cycle.
   --  Set FINISHED to true if this is the last cycle.
   procedure Simulation_Cycle (Finished : out Boolean)
   is
      Has_Run : Boolean;
      Tn : Std_Time;
   begin
      --  LRM93 12.6.4
      --  A simulation cycle consists of the following steps:
      --
      --  a) The current time, Tc is set equal to Tn.  Simulation is complete
      --     when Tn = TIME'HIGH and there are no active drivers or process
      --     resumptions at Tn.
      --  GHDL: this is done at the last step of the cycle.
      null;

      --  b) Each active explicit signal in the model is updated.  (Events
      --     may occur on signals as a result).
      --  c) Each implicit signal in the model is updated.  (Events may occur
      --     on signals as a result.)
      Update_Signals;

      --  d) For each process P, if P is currently sensitive to a signal S and
      --     if an event has occured on S in this simulation cycle, then P
      --     resumes.
      for I in Process_Table.First .. Process_Table.Last loop
         declare
            Proc : Process_Type renames Process_Table.Table (I);
            El : Sensitivity_Acc;
         begin
            case Proc.State is
               when State_Sensitized =>
                  null;
               when State_Wait =>
                  if Proc.Timeout = Current_Time then
                     Proc.Timeout := Bad_Time;
                     Proc.Resumed := True;
                     Proc.State := State_Timeout;
                  else
                     El := Proc.Sensitivity;
                     while El /= null loop
                        if El.Sig.Event then
                           Proc.Resumed := True;
                           exit;
                        else
                           El := El.Next;
                        end if;
                     end loop;
                  end if;
               when State_Timeout =>
                  Internal_Error ("process in timeout");
               when State_Dead =>
                  null;
            end case;
         end;
      end loop;

      --  e) Each nonpostponed that has resumed in the current simulation cycle
      --     is executed until it suspends.
      Run_Processes (Postponed => False, Has_Run => Has_Run);

      --  f) The time of the next simulation cycle, Tn, is determined by
      --     setting it to the earliest of
      --     1) TIME'HIGH
      --     2) The next time at which a driver becomes active, or
      --     3) The next time at which a process resumes.
      --     If Tn = Tc, then the next simulation cycle (if any) will be a
      --     delta cycle.
      Tn := Compute_Next_Time;

      --  g) If the next simulation cycle will be a delta cycle, the remainder
      --     of the step is skipped.
      --     Otherwise, each postponed process that has resumed but has not
      --     been executed since its last resumption is executed until it
      --     suspends.  Then Tn is recalculated according to the rules of
      --     step f.  It is an error if the execution of any postponed
      --     process causes a delta cycle to occur immediatly after the
      --     current simulation cycle.
      Finished := False;
      if Tn = Current_Time then
         if Current_Time = Last_Time and then not Has_Run then
            Finished := True;
         else
            Current_Delta := Current_Delta + 1;
         end if;
      else
         Current_Delta := 0;
         Run_Processes (Postponed => True, Has_Run => Has_Run);
         Tn := Compute_Next_Time;
         if Tn = Current_Time then
            Error ("postponed process causes a delta cycle");
         end if;
         Current_Time := Tn;
      end if;
   end Simulation_Cycle;

   procedure Start
   is
      Finished : Boolean;
   begin
      --Put_Line ("grt.processes:" & Process_Id'Image (Process_Table.Last)
      --          & " process(es)");

      if Options.Disp_Design_Tree then
         --Grt.Disp.Disp_Hierarchy;
         Grt.Disp.Disp_Hierarchy_Tree;
      end if;
      if Options.Disp_Signals_Order then
         Grt.Disp.Disp_Signals_Order;
      end if;
      if Options.Disp_Sig_Types then
         Grt.Disp.Disp_Signals_Type;
      end if;

      Grt.Vcd.Vcd_Start;

      Initialization_Phase;
      Current_Delta := 0;
      if Options.Trace_Signals then
         Grt.Disp.Disp_Simple_Signals (Options.Disp_Sources);
      end if;

      loop
         if Options.Disp_Time then
            Grt.Disp.Disp_Now;
         end if;
         Simulation_Cycle (Finished);
         if Options.Trace_Signals then
            Grt.Disp.Disp_Simple_Signals (Options.Disp_Sources);
         end if;
         exit when Finished;
         if Current_Delta = 0 then
            Grt.Vcd.Vcd_Cycle;
         end if;
         if Current_Delta >= Grt.Options.Stop_Delta then
            Info ("simulation stopped by --stop-delta");
            exit;
         end if;
         if Current_Time > Grt.Options.Stop_Time then
            Info ("simulation stopped by --stop-time");
            exit;
         end if;
      end loop;

      Grt.Vcd.Vcd_End;

   end Start;

end Grt.Processes;
