with Types; use Types;
with Tokens; use Tokens;

package Scan is
   -- Global variables
   -- The token that was just scanned.
   -- When the token was eaten, you can call invalidate_current_token to
   -- set it to tok_invalid.
   -- Current_token should not be written outside of scan package.
   -- It can be replaced by a function call.
   Current_Token: Token_Type := Tok_Invalid;

   -- Simply set current_token to tok_invalid.
   procedure Invalidate_Current_Token;
   pragma Inline (Invalidate_Current_Token);

   -- When CURRENT_TOKEN is an tok_identifier, tok_char or tok_string,
   -- its name_id can be got via this function.
   function Current_Identifier return Name_Id;
   pragma Inline (Current_Identifier);

   -- Set Current_identifier to null_identifier.
   -- Can be used to catch bugs.
   procedure Invalidate_Current_Identifier;
   pragma Inline (Invalidate_Current_Identifier);

   -- When CURRENT_TOKEN is tok_integer, returns the value.
   -- When CURRENT_TOKEN is tok_bit_string, returns the base.
   function Current_Iir_Int64 return Iir_Int64;
   pragma Inline (Current_Iir_Int64);

   -- When CURRENT_TOKEN is tok_real, it returns the value.
   function Current_Iir_Fp64 return Iir_Fp64;
   pragma Inline (Current_Iir_Fp64);

   -- Advances the lexical analyser.  Put a new token into current_token.
   procedure Scan;

   -- Initialize the scanner with file FILENAME in DIRECTORY.
   --  SUCCESS id set to false if the file has not been opened.
   procedure Set_File (Source_File : Source_File_Entry;
                       Success : out Boolean);
   procedure Set_File (Directory : Name_Id;
                       Filename: Name_Id;
                       Success : out Boolean);

   procedure Set_Current_Position (Position: Source_Ptr);

   -- Finalize the scanner.
   procedure Close_File;

   -- Get the current location, or the location of the current token.
   -- Since a token cannot spread over lines, file and line of the current
   -- token are the same as those of the current position.
   function Get_Current_File return Name_Id;
   function Get_Current_Source_File return Source_File_Entry;
   function Get_Current_Line return Natural;
   function Get_Current_Column return Natural;
   function Get_Token_Location return Location_Type;
   function Get_Token_Column return Natural;
   function Get_Token_Position return Source_Ptr;

   -- The context contains the whole internal state of the scanner, ie
   -- it can be used to push/pop a lexical analysis, to restart the
   -- scanner from a context marking a previous point.
   type Scan_Context is limited private;

   -- These two functions can be used together to do backtracking.
   procedure Scan_Save_Context (Context: out Scan_Context);
   procedure Scan_Restore_Context (Context: in Scan_Context);

   -- These two functions can be used to do a recursive parser.
   -- After a suspend, no more tokens can be get with the scan function, the
   --  lexical analyser is not active anymore.
   -- A resume can be only used when the lexical analyser is not initialized,
   -- to restart a suspended analysis.
   -- Use set_file or scan_resume.
   procedure Scan_Suspend (Context: out Scan_Context);
   procedure Scan_Resume (Context: in Scan_Context);

private
   type Scan_Context is record
         Source: File_Buffer_Acc;
         Source_File: Source_File_Entry;
         Line_Number: Natural;
         Line_Pos: Source_Ptr;
         Pos: Source_Ptr;
         Token_Pos: Source_Ptr;
         File_Len: Integer;
         File_Name: Name_Id;
         Token: Token_Type;
         Prev_Token: Token_Type;
         Identifier: Name_Id;
         Int64: Iir_Int64;
         Fp64: Iir_Fp64;
   end record;
end Scan;
